@@ -1,4 +1,4 @@
-// Copyright 2018 Chia Network Inc
+// Copyright 2020 Chia Network Inc
 
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
@@ -26,66 +26,23 @@ namespace py = pybind11;
 using namespace bls;
 using std::vector;
 
-template <typename... Args>
-using overload_cast_ = pybind11::detail::overload_cast_impl<Args...>;
+
+/* This class releases the Python GIL until the end of the scope.
+   This is different from gil_scoped_release in that it can't  be reacquired early. */
+// class PythonGIL {
+// public:
+//     PythonGIL() { _save = PyEval_SaveThread(); }
+//     ~PythonGIL() { PyEval_RestoreThread(_save); }
+// private:
+//     PyThreadState *_save;
+// };
 
 PYBIND11_MODULE(blspy, m)
 {
-    py::class_<BNWrapper>(m, "BNWrapper")
-        .def(py::init(&BNWrapper::FromByteVector))
-        .def(py::init([](py::int_ pyint) {
-            size_t n_bytes = 1 + ((_PyLong_NumBits(pyint.ptr()) + 7) / 8);
-            std::vector<uint8_t> buffer(n_bytes, 0);
-            if (_PyLong_AsByteArray(
-                    (PyLongObject *)pyint.ptr(), buffer.data(), n_bytes, 0, 0) <
-                0) {
-                throw std::invalid_argument("Failed to cast int to BNWrapper");
-            }
-            return BNWrapper::FromByteVector(buffer);
-        }))
-        .def(
-            "__repr__",
-            [](const BNWrapper &b) {
-                std::stringstream s;
-                s << b;
-                return "<BNWrapper " + s.str() + ">";
-            })
-        .def(
-            "__deepcopy__",
-            [](const BNWrapper &k, const py::object &memo) {
-                return BNWrapper(k);
-            })
-        .def(
-            "__str__",
-            [](const BNWrapper &b) {
-                std::stringstream s;
-                s << b;
-                return s.str();
-            })
-        .def("__bytes__", [](const BNWrapper &b) {
-            std::stringstream s;
-            int length = bn_size_bin(*b.b);
-            uint8_t *out = new uint8_t[length];
-            bn_write_bin(out, length, *b.b);
-            s << out;
-            delete[] out;
-            return py::bytes(s.str());
-        });
-
-    py::implicitly_convertible<py::int_, BNWrapper>();
-
     py::class_<PrivateKey>(m, "PrivateKey")
         .def_property_readonly_static(
             "PRIVATE_KEY_SIZE",
             [](py::object self) { return PrivateKey::PRIVATE_KEY_SIZE; })
-        .def(
-            "from_seed",
-            [](const py::bytes &b) {
-                std::string str(b);
-                const uint8_t *input =
-                    reinterpret_cast<const uint8_t *>(str.data());
-                return PrivateKey::FromSeed(input, len(b));
-            })
         .def(
             "from_bytes",
             [](py::buffer const b) {
@@ -99,23 +56,13 @@ PYBIND11_MODULE(blspy, m)
                         "Length of bytes object not equal to PrivateKey::SIZE");
                 }
                 auto data_ptr = reinterpret_cast<const uint8_t *>(info.ptr);
-                return PrivateKey::FromBytes(data_ptr);
+                // PythonGIL release_lock;
+                return PrivateKey::FromBytes(Bytes(data_ptr, PrivateKey::PRIVATE_KEY_SIZE));
             })
         .def(
             "__bytes__",
             [](const PrivateKey &k) {
-                uint8_t *output =
-                    Util::SecAlloc<uint8_t>(PrivateKey::PRIVATE_KEY_SIZE);
-                k.Serialize(output);
-                py::bytes ret = py::bytes(
-                    reinterpret_cast<char *>(output),
-                    PrivateKey::PRIVATE_KEY_SIZE);
-                Util::SecFree(output);
-                return ret;
-            })
-        .def(
-            "serialize",
-            [](const PrivateKey &k) {
+                // PythonGIL release_lock;
                 uint8_t *output =
                     Util::SecAlloc<uint8_t>(PrivateKey::PRIVATE_KEY_SIZE);
                 k.Serialize(output);
@@ -130,18 +77,11 @@ PYBIND11_MODULE(blspy, m)
             [](const PrivateKey &k, const py::object &memo) {
                 return PrivateKey(k);
             })
-        .def("get_g1", [](const PrivateKey &k) { return k.GetG1Element(); })
-        .def(
-            "__deepcopy__",
-            [](const PrivateKey &k, const py::object &memo) {
-                return PrivateKey(k);
+        .def("get_g1", [](const PrivateKey &k) {
+                // PythonGIL release_lock; 
+                return k.GetG1Element();
             })
-        .def(
-            "derive_child",
-            [](PrivateKey &k, uint32_t index) {
-                return HDKeys::DeriveChildSk(k, index);
-            })
-
+        .def("aggregate", &PrivateKey::Aggregate)
         .def(py::self == py::self)
         .def(py::self != py::self)
         .def("__repr__", [](const PrivateKey &k) {
@@ -165,17 +105,34 @@ PYBIND11_MODULE(blspy, m)
     });
 
     py::class_<BasicSchemeMPL>(m, "BasicSchemeMPL")
-        .def("sk_to_g1", &BasicSchemeMPL::SkToG1)
+        .def("sk_to_g1", [](const PrivateKey &seckey){ return BasicSchemeMPL().SkToG1(seckey); })
         .def(
-            "aggregate",
-            overload_cast_<const vector<G2Element> &>()(
-                &BasicSchemeMPL::Aggregate))
+            "key_gen",
+            [](const py::bytes &b) {
+                std::string str(b);
+                const uint8_t *input =
+                    reinterpret_cast<const uint8_t *>(str.data());
+                const vector<uint8_t> inputVec(input, input + len(b));
+                return BasicSchemeMPL().KeyGen(inputVec);
+            })
+        .def("derive_child_sk", [](const PrivateKey& sk, uint32_t index){
+            return BasicSchemeMPL().DeriveChildSk(sk, index);
+        })
+        .def("derive_child_sk_unhardened", [](const PrivateKey& sk, uint32_t index){
+            return BasicSchemeMPL().DeriveChildSkUnhardened(sk, index);
+        })
+        .def("derive_child_pk_unhardened", [](const G1Element& pk, uint32_t index){
+            return BasicSchemeMPL().DeriveChildPkUnhardened(pk, index);
+        })
+        .def("aggregate", [](const vector<G2Element> &signatures) {
+            return BasicSchemeMPL().Aggregate(signatures);
+        })
         .def(
             "sign",
             [](const PrivateKey &pk, const py::bytes &msg) {
                 std::string s(msg);
                 vector<uint8_t> v(s.begin(), s.end());
-                return BasicSchemeMPL::SignNative(pk, v);
+                return BasicSchemeMPL().Sign(pk, v);
             })
         .def(
             "verify",
@@ -183,11 +140,12 @@ PYBIND11_MODULE(blspy, m)
                const py::bytes &msg,
                const G2Element &sig) {
                 std::string s(msg);
+                // PythonGIL release_lock;
                 vector<uint8_t> v(s.begin(), s.end());
-                return BasicSchemeMPL::Verify(pk, v, sig);
+                return BasicSchemeMPL().Verify(pk, v, sig);
             })
         .def(
-            "agg_verify",
+            "aggregate_verify",
             [](const vector<G1Element> &pks,
                const vector<py::bytes> &msgs,
                const G2Element &sig) {
@@ -196,22 +154,51 @@ PYBIND11_MODULE(blspy, m)
                     std::string s(msgs[i]);
                     vecs[i] = vector<uint8_t>(s.begin(), s.end());
                 }
-
-                return BasicSchemeMPL::AggregateVerify(pks, vecs, sig);
+                // PythonGIL release_lock;
+                return BasicSchemeMPL().AggregateVerify(pks, vecs, sig);
+            })
+        .def(
+            "g2_from_message",
+            [](const py::bytes &msg) {
+                const auto msg_str = std::string(msg);
+                const auto msg_bytes = Bytes((const uint8_t *)msg_str.c_str(), msg_str.size());
+                return G2Element::FromMessage(
+                    msg_bytes,
+                    (const uint8_t *)BasicSchemeMPL::CIPHERSUITE_ID.c_str(),
+                    BasicSchemeMPL::CIPHERSUITE_ID.size()
+                );
             });
 
     py::class_<AugSchemeMPL>(m, "AugSchemeMPL")
-        .def("sk_to_g1", &AugSchemeMPL::SkToG1)
+        .def("sk_to_g1", [](const PrivateKey &seckey){ return AugSchemeMPL().SkToG1(seckey); })
         .def(
-            "aggregate",
-            overload_cast_<const vector<G2Element> &>()(
-                &AugSchemeMPL::Aggregate))
+            "key_gen",
+            [](const py::bytes &b) {
+                std::string str(b);
+                const uint8_t *input =
+                    reinterpret_cast<const uint8_t *>(str.data());
+                const vector<uint8_t> inputVec(input, input + len(b));
+                return AugSchemeMPL().KeyGen(inputVec);
+            })
+        .def("derive_child_sk", [](const PrivateKey& sk, uint32_t index){
+            // PythonGIL release_lock;
+            return AugSchemeMPL().DeriveChildSk(sk, index);
+        })
+        .def("derive_child_sk_unhardened", [](const PrivateKey& sk, uint32_t index){
+            return AugSchemeMPL().DeriveChildSkUnhardened(sk, index);
+        })
+        .def("derive_child_pk_unhardened", [](const G1Element& pk, uint32_t index){
+            return AugSchemeMPL().DeriveChildPkUnhardened(pk, index);
+        })
+        .def("aggregate", [](const vector<G2Element>& signatures) {
+            return AugSchemeMPL().Aggregate(signatures);
+        })
         .def(
             "sign",
             [](const PrivateKey &pk, const py::bytes &msg) {
                 std::string s(msg);
                 vector<uint8_t> v(s.begin(), s.end());
-                return AugSchemeMPL::SignNative(pk, v);
+                return AugSchemeMPL().Sign(pk, v);
             })
         .def(
             "sign",
@@ -220,7 +207,7 @@ PYBIND11_MODULE(blspy, m)
                const G1Element &prepend_pk) {
                 std::string s(msg);
                 vector<uint8_t> v(s.begin(), s.end());
-                return AugSchemeMPL::SignNative(pk, v, prepend_pk);
+                return AugSchemeMPL().Sign(pk, v, prepend_pk);
             })
         .def(
             "verify",
@@ -228,11 +215,12 @@ PYBIND11_MODULE(blspy, m)
                const py::bytes &msg,
                const G2Element &sig) {
                 std::string s(msg);
+                // PythonGIL release_lock;
                 vector<uint8_t> v(s.begin(), s.end());
-                return AugSchemeMPL::Verify(pk, v, sig);
+                return AugSchemeMPL().Verify(pk, v, sig);
             })
         .def(
-            "agg_verify",
+            "aggregate_verify",
             [](const vector<G1Element> &pks,
                const vector<py::bytes> &msgs,
                const G2Element &sig) {
@@ -241,26 +229,50 @@ PYBIND11_MODULE(blspy, m)
                     std::string s(msgs[i]);
                     vecs[i] = vector<uint8_t>(s.begin(), s.end());
                 }
-
-                return AugSchemeMPL::AggregateVerify(pks, vecs, sig);
+                // PythonGIL release_lock;
+                return AugSchemeMPL().AggregateVerify(pks, vecs, sig);
+            })
+        .def(
+            "g2_from_message",
+            [](const py::bytes &msg) {
+                const auto msg_str = std::string(msg);
+                const auto msg_bytes = Bytes((const uint8_t *)msg_str.c_str(), msg_str.size());
+                return G2Element::FromMessage(
+                    msg_bytes,
+                    (const uint8_t *)AugSchemeMPL::CIPHERSUITE_ID.c_str(),
+                    AugSchemeMPL::CIPHERSUITE_ID.size()
+                );
             });
 
     py::class_<PopSchemeMPL>(m, "PopSchemeMPL")
-        .def("sk_to_g1", &PopSchemeMPL::SkToG1)
+        .def("sk_to_g1", [](const PrivateKey &seckey){ return PopSchemeMPL().SkToG1(seckey); })
         .def(
-            "aggregate",
-            overload_cast_<const vector<vector<uint8_t>> &>()(
-                &PopSchemeMPL::Aggregate))
-        .def(
-            "aggregate",
-            overload_cast_<const vector<G2Element> &>()(
-                &PopSchemeMPL::Aggregate))
+            "key_gen",
+            [](const py::bytes &b) {
+                std::string str(b);
+                const uint8_t *input =
+                    reinterpret_cast<const uint8_t *>(str.data());
+                const vector<uint8_t> inputVec(input, input + len(b));
+                return PopSchemeMPL().KeyGen(inputVec);
+            })
+        .def("derive_child_sk", [](const PrivateKey& sk, uint32_t index){
+            return PopSchemeMPL().DeriveChildSk(sk, index);
+        })
+        .def("derive_child_sk_unhardened", [](const PrivateKey& sk, uint32_t index){
+            return PopSchemeMPL().DeriveChildSkUnhardened(sk, index);
+        })
+        .def("derive_child_pk_unhardened", [](const G1Element& pk, uint32_t index){
+            return PopSchemeMPL().DeriveChildPkUnhardened(pk, index);
+        })
+        .def("aggregate", [](const vector<G2Element>& signatures) {
+            return PopSchemeMPL().Aggregate(signatures);
+        })
         .def(
             "sign",
             [](const PrivateKey &pk, const py::bytes &msg) {
                 std::string s(msg);
                 vector<uint8_t> v(s.begin(), s.end());
-                return PopSchemeMPL::SignNative(pk, v);
+                return PopSchemeMPL().Sign(pk, v);
             })
         .def(
             "verify",
@@ -268,11 +280,12 @@ PYBIND11_MODULE(blspy, m)
                const py::bytes &msg,
                const G2Element &sig) {
                 std::string s(msg);
+                // PythonGIL release_lock;
                 vector<uint8_t> v(s.begin(), s.end());
-                return PopSchemeMPL::Verify(pk, v, sig);
+                return PopSchemeMPL().Verify(pk, v, sig);
             })
         .def(
-            "agg_verify",
+            "aggregate_verify",
             [](const vector<G1Element> &pks,
                const vector<py::bytes> &msgs,
                const G2Element &sig) {
@@ -281,34 +294,42 @@ PYBIND11_MODULE(blspy, m)
                     std::string s(msgs[i]);
                     vecs[i] = vector<uint8_t>(s.begin(), s.end());
                 }
-
-                return PopSchemeMPL::AggregateVerify(pks, vecs, sig);
+                // PythonGIL release_lock;
+                return PopSchemeMPL().AggregateVerify(pks, vecs, sig);
             })
-        .def("pop_prove", &PopSchemeMPL::PopProveNative)
         .def(
-            "pop_verify",
-            overload_cast_<const G1Element &, const G2Element &>()(
-                &PopSchemeMPL::PopVerify))
+            "g2_from_message",
+            [](const py::bytes &msg) {
+                const auto msg_str = std::string(msg);
+                const auto msg_bytes = Bytes((const uint8_t *)msg_str.c_str(), msg_str.size());
+                return G2Element::FromMessage(
+                    msg_bytes,
+                    (const uint8_t *)PopSchemeMPL::CIPHERSUITE_ID.c_str(),
+                    PopSchemeMPL::CIPHERSUITE_ID.size()
+                );
+            })
+        .def("pop_prove", [](const PrivateKey& privateKey){
+            return PopSchemeMPL().PopProve(privateKey);
+        })
+        .def("pop_verify", [](const G1Element& pubkey, const G2Element& signature){
+            // PythonGIL release_lock;
+            return PopSchemeMPL().PopVerify(pubkey, signature);
+        })
         .def(
-            "fast_agg_verify",
-            overload_cast_<
-                const vector<G1Element> &,
-                const vector<uint8_t> &,
-                const G2Element &>()(&PopSchemeMPL::FastAggregateVerify))
-        .def(
-            "fast_agg_verify",
+            "fast_aggregate_verify",
             [](const vector<G1Element> &pks,
                const py::bytes &msg,
                const G2Element &sig) {
                 std::string s(msg);
+                // PythonGIL release_lock;
                 vector<uint8_t> v(s.begin(), s.end());
-                return PopSchemeMPL::FastAggregateVerify(pks, v, sig);
+                return PopSchemeMPL().FastAggregateVerify(pks, v, sig);
             });
 
     py::class_<G1Element>(m, "G1Element")
         .def_property_readonly_static(
             "SIZE", [](py::object self) { return G1Element::SIZE; })
-        .def(py::init<>())
+        .def(py::init([](){ return G1Element();}))
         .def(py::init(&G1Element::FromByteVector))
         .def(py::init([](py::int_ pyint) {
             std::vector<uint8_t> buffer(G1Element::SIZE, 0);
@@ -349,12 +370,13 @@ PYBIND11_MODULE(blspy, m)
                         "Length of bytes object not equal to G1Element::SIZE");
                 }
                 auto data_ptr = reinterpret_cast<const uint8_t *>(info.ptr);
-                return G1Element::FromBytes(data_ptr);
+                // PythonGIL release_lock;
+                return G1Element::FromBytes(Bytes(data_ptr, G1Element::SIZE));
             })
         .def("generator", &G1Element::Generator)
-        .def("from_message", &G1Element::FromMessage)
-        .def("pair", &G1Element::pair)
-        .def("inverse", &G1Element::Inverse)
+        .def("from_message", py::overload_cast<const std::vector<uint8_t>&, const uint8_t*, int>(&G1Element::FromMessage))
+        .def("pair", &G1Element::Pair)
+        .def("negate", &G1Element::Negate)
         .def("get_fingerprint", &G1Element::GetFingerprint)
 
         .def(py::self == py::self)
@@ -366,27 +388,9 @@ PYBIND11_MODULE(blspy, m)
             })
         .def(
             "__add__",
-            [](G1Element &self, G1Element &other) { return self + other; },
-            py::is_operator())
-        .def(
-            "__iadd__",
             [](G1Element &self, G1Element &other) {
-                self += other;
-                return self;
-            },
-            py::is_operator())
-        .def(
-            "__imul__",
-            [](G1Element &self, bn_t other) {
-                self *= (*(bn_t *)&other);
-                return self;
-            },
-            py::is_operator())
-        .def(
-            "__imul__",
-            [](G1Element &self, BNWrapper other) {
-                self *= (*other.b);
-                return self;
+                // PythonGIL release_lock;
+                return self + other;
             },
             py::is_operator())
         .def(
@@ -395,20 +399,12 @@ PYBIND11_MODULE(blspy, m)
                 return self * (*(bn_t *)&other);
             },
             py::is_operator())
-        .def(
-            "__mul__",
-            [](G1Element &self, BNWrapper other) { return self * (*other.b); },
-            py::is_operator())
         .def(
             "__rmul__",
             [](G1Element &self, bn_t other) {
                 return self * (*(bn_t *)&other);
             },
             py::is_operator())
-        .def(
-            "__rmul__",
-            [](G1Element &self, BNWrapper other) { return self * (*other.b); },
-            py::is_operator())
         .def(
             "__and__",
             [](G1Element &self, G2Element &other) { return self & other; },
@@ -430,11 +426,12 @@ PYBIND11_MODULE(blspy, m)
         .def(
             "__bytes__",
             [](const G1Element &ele) {
-                uint8_t *out = new uint8_t[G1Element::SIZE];
-                ele.Serialize(out);
+                vector<uint8_t> out;
+                // Py_BEGIN_ALLOW_THREADS
+                out = ele.Serialize();
+                // Py_END_ALLOW_THREADS
                 py::bytes ans = py::bytes(
-                    reinterpret_cast<const char *>(out), G1Element::SIZE);
-                delete[] out;
+                    reinterpret_cast<const char *>(out.data()), G1Element::SIZE);
                 return ans;
             })
         .def("__deepcopy__", [](const G1Element &ele, const py::object &memo) {
@@ -444,7 +441,7 @@ PYBIND11_MODULE(blspy, m)
     py::class_<G2Element>(m, "G2Element")
         .def_property_readonly_static(
             "SIZE", [](py::object self) { return G2Element::SIZE; })
-        .def(py::init<>())
+        .def(py::init([](){ return G2Element();}))
         .def(py::init(&G2Element::FromByteVector))
         .def(py::init([](py::buffer const b) {
             py::buffer_info info = b.request();
@@ -485,12 +482,12 @@ PYBIND11_MODULE(blspy, m)
                         "Length of bytes object not equal to G2Element::SIZE");
                 }
                 auto data_ptr = reinterpret_cast<const uint8_t *>(info.ptr);
-                return G2Element::FromBytes(data_ptr);
+                return G2Element::FromBytes(Bytes(data_ptr, G2Element::SIZE));
             })
         .def("generator", &G2Element::Generator)
-        .def("from_message", &G2Element::FromMessage)
-        .def("pair", &G2Element::pair)
-        .def("inverse", &G2Element::Inverse)
+        .def("from_message", py::overload_cast<const std::vector<uint8_t>&, const uint8_t*, int>(&G2Element::FromMessage))
+        .def("pair", &G2Element::Pair)
+        .def("negate", &G2Element::Negate)
         .def(
             "__deepcopy__",
             [](const G2Element &g2, const py::object &memo) {
@@ -503,48 +500,18 @@ PYBIND11_MODULE(blspy, m)
             "__add__",
             [](G2Element &self, G2Element &other) { return self + other; },
             py::is_operator())
-        .def(
-            "__iadd__",
-            [](G2Element &self, G2Element &other) {
-                self += other;
-                return self;
-            },
-            py::is_operator())
-        .def(
-            "__imul__",
-            [](G2Element &self, bn_t other) {
-                self *= (*(bn_t *)&other);
-                return self;
-            },
-            py::is_operator())
-        .def(
-            "__imul__",
-            [](G2Element &self, BNWrapper other) {
-                self *= (*other.b);
-                return self;
-            },
-            py::is_operator())
-
         .def(
             "__mul__",
             [](G2Element &self, bn_t other) {
                 return self * (*(bn_t *)&other);
             },
             py::is_operator())
-        .def(
-            "__mul__",
-            [](G2Element &self, BNWrapper other) { return self * (*other.b); },
-            py::is_operator())
         .def(
             "__rmul__",
             [](G2Element &self, bn_t other) {
                 return self * (*(bn_t *)&other);
             },
             py::is_operator())
-        .def(
-            "__rmul__",
-            [](G2Element &self, BNWrapper other) { return self * (*other.b); },
-            py::is_operator())
 
         .def(
             "__repr__",
@@ -563,11 +530,9 @@ PYBIND11_MODULE(blspy, m)
         .def(
             "__bytes__",
             [](const G2Element &ele) {
-                uint8_t *out = new uint8_t[G2Element::SIZE];
-                ele.Serialize(out);
+                vector<uint8_t> out = ele.Serialize();
                 py::bytes ans = py::bytes(
-                    reinterpret_cast<const char *>(out), G2Element::SIZE);
-                delete[] out;
+                    reinterpret_cast<const char *>(out.data()), G2Element::SIZE);
                 return ans;
             })
         .def("__deepcopy__", [](const G2Element &ele, const py::object &memo) {
@@ -577,7 +542,6 @@ PYBIND11_MODULE(blspy, m)
     py::class_<GTElement>(m, "GTElement")
         .def_property_readonly_static(
             "SIZE", [](py::object self) { return GTElement::SIZE; })
-        .def(py::init<>())
         .def(py::init(&GTElement::FromByteVector))
         .def(py::init([](py::buffer const b) {
             py::buffer_info info = b.request();
@@ -618,8 +582,9 @@ PYBIND11_MODULE(blspy, m)
                         "Length of bytes object not equal to GTElement::SIZE");
                 }
                 auto data_ptr = reinterpret_cast<const uint8_t *>(info.ptr);
-                return GTElement::FromBytes(data_ptr);
+                return GTElement::FromBytes(Bytes(data_ptr, GTElement::SIZE));
             })
+        .def("unity", &GTElement::Unity)
         .def(py::self == py::self)
         .def(py::self != py::self)
         .def(
@@ -651,6 +616,12 @@ PYBIND11_MODULE(blspy, m)
                 delete[] out;
                 return ans;
             })
+        .def(
+            "__mul__",
+            [](GTElement &self, GTElement &other) {
+                return self * other;
+            },
+            py::is_operator())
         .def("__deepcopy__", [](const GTElement &ele, const py::object &memo) {
             return GTElement(ele);
         });
